#!/usr/bin/perl
#
# lt24merge.pl - Merge LED Toy firmware with images
#
# Written 2005 by Werner Almesberger
# Copyright 2005 by Werner Almesberger
#

#
# PIC memory
#
$MAX_BYTES = (4096-0x100)*2;


#
# n-th diode to port and phase:
#
# Value Port Phase
#  0- 7  B     1
#  8-15  A     1
# 16-23  B     2
# 24-31  A     2
#

#
# The 12 pixels in a group, which are spread out over 16 ports, are packed
# into a 14 bit value, together with the phase select bits.
#
# LED     Port  Encoding  Comment      
#  0,  1  RB1   1+8      
#  2,  3  RB3   3+8  
#  4,  5  RB2   2+8
#  6,  7  RB6   5         instead of unused RA5
#  8,  9  RB4   4+8
# 10, 11  RB7   5+8       instead of unused RB5       
# 12, 13  RA6   6   
# 14, 15  RA7   7
# 16, 17  RA0   0
# 18, 19  RA1   1
# 20, 21  RA2   2                
# 22, 23  RA3   3                  
#

#
# The original "A" map
#

@map =
  ( 1+ 8, 1+24, 3+ 8, 3+24, 2+ 8, 2+24, 5,    5+16,
    4+ 8, 4+24, 5+ 8, 5+24, 6,    6+16, 7,    7+16,
    0,    0+16, 1,    1+16, 2,    2+16, 3,    3+16 );
$xor = 0;


sub swap
{
    local ($a,$b) = @_;

    ($map[$a],$map[$a+1],$map[$b],$map[$b+1]) =
      ($map[$b],$map[$b+1],$map[$a],$map[$a+1]);
}


#
# RA4 selects phase 1, RB0 selects phase 2
#

$PHASE_SELECT = (1 << 4) | (1 << (0+24));


sub convert
{
    local $width = shift(@_);
    local $res = "";

    @v = split(/,\s*/,$_[0]);
    map { $_ = hex $_; } @v;
    for ($x = 0; $x != $width; $x++) {
	$pat = $PHASE_SELECT;
	for ($y = 0; $y != 24; $y++) {
	    $pix = $x+$y*(($width+7) & ~7);
	    $pat |= 1 << $map[$y] if $v[$pix >> 3] & (1 << ($pix & 7));
	}
	$res .= pack("V",$pat ^ $xor);
    }
    return $res;
}


sub load_xbm
{
    local ($s,$x,$bytes);

    open(FILE,$_) || die "$_[0]: $!";
    $s = join("",<FILE>);
    close FILE;
    $s =~ /#define\s\S+_width\s(\d+)\s+#define\s\S+_height\s24\s+/ ||
      die "$_[0]: not a valid X bitmap file";
    die "image width $1 out of range [1, 255]" if $1 < 1 || $1 > 255;
    $x = $1;
    $bytes = 3*(($x+7) & ~7)-1;
    $' =~ /^{0}static\s+(unsigned\s+)*char\s+\S+_bits\[\]\s=\s\{\s+
{0}(0x[0-9a-f]{2}(,\s*0x[0-9a-f]{2}){$bytes})\};\s*$/ ||
      die "$_[0]: not a valid X bitmap file";
    return pack("v",$x == 24 ? 0 : 0x100+$x).&convert($x,$2);
}


sub load_seq
{
}


$fw = shift(@ARGV);
if (!defined $fw) {
    print STDERR "usage: $ARGV[0] firmware.bin [image.xbm ...]\n";
    exit(1);
}

if ($fw =~ /-LB24A\.bin$/) {
    # Nothing to do
}
elsif ($fw =~ /-LB24(C|V[01])\.bin$/) {
    #
    # V0 map is like A map, but with RB4 and RB6 swapped
    #
    &swap(6,8);
}
else {
    #
    # V2 map is like V0, but with RB2 and RB3 swapped and all polarities
    # reversed
    #
    print STDERR "warning: unknown firmware version, assuming V2 compatibility"
      if $fw !~ /-LB24V[23]\.bin$/;
    &swap(6,8);
    &swap(2,4);
    $xor = 0x3fff3fff;
}


open(FILE,$fw) || die "$fw: $!";
while (<FILE>) {
    $bin .= $_;
}
close FILE;

$table = "";
$data = "";
for (@ARGV) {
    $table .= pack("v",@ARGV+1+(length($data)+length($bin))/2+0x100);
    $data .= &load_xbm($_);
}

$bin .= $table.pack("v",0).$data;

$len = length($bin);
if ($len > $MAX_BYTES) {
    print STDERR "output is too big ($len > $MAX_BYTES)\n";
    exit(1);
}

print $bin;
