;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; LED Toy, Control Program
;
; Written 2005 by Werner Almesberger
; Copyright 2005 by Werner Almesberger
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


	title	"LED Toy"

	list	P = 16F88

	include	"p16f88.inc"

config = _CP_OFF & _DEBUG_OFF & _WRT_PROTECT_256 & _CPD_OFF
config	= config & _LVP_OFF & _BODEN_OFF & _MCLR_OFF & _PWRTE_ON & _WDT_OFF
config	= config & _INTRC_IO

	__config	config

; PICP mis-programs the 16F88. Therefore, we pretend it's a 16F819.
; Unfortunately, the 16F819 doesn't have a CONFIG2 word, so we can't define it
; here.
;
;	__config        _CONFIG2,_IESO_OFF & _FCMEN_OFF

	radix	dec
	errorlevel -302		; yes, I DO make sure RP0 has the right value


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Translate board version to hardware properties    
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Defaults are for LB24V3

HW_SINK		= 1	; PIC sinks current
HW_SWAP_RB2_RB3 = 1	; swap RB2 and RB3
HW_SWAP_RB4_RB6 = 1	; swap RB4 and RB6

ifdef LB24A  
HW_SINK		= 0
HW_SWAP_RB2_RB3 = 0
HW_SWAP_RB4_RB6 = 0
endif

ifdef LB24C
#define LB24V1		; "C" revision is identical to V1
endif

ifdef LB24V0
#define LB24V1		; V0 is identical to V1
endif

ifdef LB24V1
HW_SINK		= 0
HW_SWAP_RB2_RB3 = 0
endif

; V2 is identical to V3


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Constants
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

; Connections

#define	BUTTON	PORTB,5		; shared with TX
#define	RX	PORTB,2		; shared with LED
#define TX	PORTB,5		; shared with button
#define PHASE0	PORTB,0		; LED 0, 2, 4, ..., 22
#define PHASE1	PORTA,4		; LED 1, 3, 5, ..., 23
#define ILLPWR	PORTA,2		; RA2
#define ILLANP	PORTA,3		; RA3
ILLAN		= 3		; RA3

; Number of LEDs

#define NUM_LEDS 24

; Current button status

#define CURRBUT	currbut,0

; Function LEDs

L_RUN		= 10
L_SELECT	= 14
L_ROTATE	= 16
L_FREQ		= 18
L_GAP		= 20
L_TO		= 22

; Timeout values (for simplicity, pretend our clock runs at 16 Hz)

T_IDLE		= 80		; "idle" timeout, 5 seconds
T_LONG		= 6		; "long" timeout, 375 ms
T_CONFIG	= 480		; "configuration idle" timeout, half a minute
T_CURRENT	= 4		; "show current" timeout, 250 ms

; Initial configuration settings

CFG_NAP		= 17		; 600 us
CFG_GAP		= 8		; 8 pixels
CFG_SHOWT	= 5		; 5 minutes

; Sequence flags

OBJ_REGULAR	= 0x00		; Object is a regular 24x24 image
OBJ_WIDE	= 0x01		; Object is a wide (24xX) image
OBJ_SEQ		= 0x02		; Object is an image sequence

SEQ_END		= 0x04		; last word of sequence data
SEQ_OFF		= 0x08		; power down
SEQ_LOOP	= 0x10		; loop back to the first image

; Miscellaneous constants

IMG_SIZE	= NUM_LEDS*2	; image size, in words
IMG_EDGE	= NUM_LEDS	; edge size

W_RUNNING	= 25		; show each LED for 49 ms
W_DEBOUNCE	= 25		; debounce for 49 ms

T0_16HZ		= 244		; 1000000/256/16 = 244.14


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Registers
; 
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


	cblock	0x20

; Configuration variables

nap				; nap duration (index 0 ... LEDs-1)
gap				; gap between images (in double pixels)
showt				; idle time (in minutes, 0 = forever)
rotate				; rotation (0 ... 3)

; Values derived fron configuration variables

napiter				; nap duration (multiples of 20 us)
showth				; idle time (in 1/16 seconds)
showtl

; Object pointer

objind				; Current object index
widepl				; Pointer to wide image, or objph = 0 if not
wideph				; a wide image
widex				; image width (only set if wide image)

; Sequence pointer

seqpl				; pointer to base of current sequence, or
seqph				; seqph = 0 if not a sequence
seqind				; index in sequence

; Timeout handler

to1h				; 1st timeout counter (counts at 16 Hz)
to1l

to2h				; 2nd timeout counter (counts at 16 Hz)
to2l

whichto				; whichto:0 is set if the 2nd timer expired

bouncing			; bouncing:0 is set for 1/T0 to debounce

; Temporary registers

ta				; delay loop counters
tb
tc

currbut				; recorded current button status

toggle				; toggle for togwth

tmp				; high-level scratch register
tmp2				; low-level scratch register

rx				; X coordinate in image rotation
ry				; Y coordinate in image rotation
rbit				; rbit:0 is value of the pixel we're moving
rtmp				; temporary data while rotating

; Local variables of "doframe"

a0				; port A in phase 0
b0				; port B in phase 0
a1				; port A in phase 1
b1				; port B in phase 1
rows				; row counter

	endc

last_global = rows		; for the monitor



	cblock	0x70		; area shared across banks

; Interrupt handler

sav_w				; saved W
sav_status			; saved STATUS

	endc


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; The program starts here
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;
; Fixed entry locations:
;
; 0   reset
; 1   polling monitor
; 2   reserved
; 3   reserved
; 4   interrupt handler
;

	org	0

	goto	start

monvec:
	goto	pollmon

	org	4

	goto	interrupt


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Device and I/O setup
;

start:

; Set internal osc to 4 MHz, so that Tcy = 1us

	bsf	STATUS,RP0
	movlw	(1 << IRCF2) | (1 << IRCF1)
	movwf	OSCCON
	; RP0 is set

; Activate pull-ups on port RB4-RB7

	; RP0 is set
	bcf	OPTION_REG,NOT_RBPU
	bcf	STATUS,RP0

; Make everything but RX an output, and drive it low, except for TX

if (HW_SINK)
	movlw	0xff		; set all outputs
	movwf	PORTA
	movwf	PORTB
else
	clrf	PORTA		; clear outputs (but set TX)
	clrf	PORTB
	bsf	TX
endif
	bsf	STATUS,RP0	; set outputs (all but RX)
	clrf	TRISA
	movlw	1 << 2
	movwf	TRISB
	bcf	STATUS,RP0

; Get rid of the ADC

	bsf	STATUS,RP0
	clrf	ANSEL
	bcf	STATUS,RP0

; Do our thing

	call	monvec		; give the monitor a chance to take over

	goto	_main		; do our thing

if (HW_SINK)

mnldon	macro
	bcf	PHASE0		; enable LEDs
	bcf	PHASE1
	bcf	PORTB,1		; light D23 and D24
	endm

; Turn the monitor LED off

mnldoff	macro
	bsf	PHASE0		; disable LEDs
	bsf	PHASE1
	bsf	PORTB,1		; turn off D23 and D24
	endm

else

mnldon	macro
	bsf	PHASE0		; enable LEDs
	bsf	PHASE1
	bsf	PORTB,1		; light D23 and D24
	endm

; Turn the monitor LED off

mnldoff	macro
	bcf	PHASE0		; disable LEDs
	bcf	PHASE1
	bcf	PORTB,1		; turn off D23 and D24
	endm

endif

	include	"monitor.inc"


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; The user interface
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


	org	256

; Fixed entry locations:
;
; 0x100  main program
; 0x101	 interrupt handler
;

_main:
	goto	main

interrupt:
	nop

	include	"ui.inc"


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Delay loop (waits approx. 3*W^3 us)
;
; Exact delay (including movlw, call and return): 3*W^3+4*W^2+4*W+6
;
; Selected delays:
;
; W   cycles	W   cycles    W   cycles    W   cycles     W    cycles
;  1       17   11     4527   21    29637   31    93347    46   300662 (300 ms)
;  2       54   12     5814   22    33974   32   102534    55   511451 (511 ms)
;  3      135   13     7325   23    38715   33   112305    69  1004853 (  1 s)
;  4      278   14     9078   24    43878   34   122678    87  2006139 (  2 s)
;  5      501   15    11091   25    49481   35   133671    99  2950503 (  3 s)
;  6      822   16    13382   26    55542   36   145302   100  3040406 (  3 s)
;  7     1259   17    15969   27    62079   37   157589   118  4985270 (  5 s)
;  8     1830   18    18870   28    69110   38   170550   149 10013253 ( 10 s)
;  9     2553   19    22103   29    76653   39   184203   255 50005251 ( 50 s)
; 10     3446   20    25686   30    84726   40   198556     0 50594822
;

waitw:
	movwf	ta
loopa:
	movwf	tb
loopb:
	movwf	tc
loopc:
	decfsz	tc,F
	goto	loopc
	decfsz	tb,F
	goto	loopb
	decfsz	ta,F
	goto	loopa
	return


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;
; Image data
;
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;


datastart:

;
; The data that follows has this structure:
;
;	dw	object0
;	dw	object1
;	...
;	dw	0
;
;	objects
;
;	fill	0,(-$ & 15)
;
; Objects:
;
; Regular image (object type 0), 24x24 pixels, 24x2 words.
;
;           +--------H--------+--------L--------+
; Word 0    | 0               | 0               |
;           +-----------------+-----------------+
; Word 2n+1 | Port B, phase 0 | Port A, phase 0 |
;           +-----------------+-----------------+
; Word 2n+2 | Port B, phase 1 | Port A, phase 1 |
;           +-----------------+-----------------+
;           :                 :                 :
;           +-----------------+-----------------+
; Word 48   | Port B, phase 1 | Port A, phase 1 |
;           +-----------------+-----------------+
;
; Note that the bit assignment is not straightforward, and some port B
; bits are folded into port A data. See lt24merge.pl for details.
;
;
; Wide image (object type 1), 24xX pixels, 1+24x2X words.
;
;           +--------H--------+--------L--------+
; Word 0    | 1               | X (columns)     |
;           +-----------------+-----------------+
; Word 2n+1 | Port B, phase 0 | Port A, phase 0 |
;           +-----------------+-----------------+
; Word 2n+2 | Port B, phase 1 | Port A, phase 1 |
;           +-----------------+-----------------+
;	    :                 :                 :
;           +-----------------+-----------------+
; Word 2X+1 | Port B, phase 1 | Port A, phase 1 |
;           +-----------------+-----------------+
;
;
; Sequence data (type 2):
;
; Each sequence begins with the type followed by the title object:
;
;           +--------H--------+--------L--------+
; Word 0    | 2               | 0               |
;           +-----------------+-----------------+
; Word 1    | Object ptr (H)  | Object ptr (L)  |
;           +-----------------+-----------------+
;
; The title object word is followed by one or more sequence image records,
; three words each:
;
;           +--------H--------+--------L--------+
; Word 0    | Flags           | Rotation        |
;           +-----------------+-----------------+
; Word 1    | Timeout (h)     | Timeout (l)     |
;           +-----------------+-----------------+
; Word 2    | Object ptr (H)  | Object ptr (L)  |
;           +-----------------+-----------------+
;
; The following flags are defined:
;
; 0x01  last of sequence, power down
; 0x02  last of sequence, loop back
;
; Only regular and wide images can be used as objects in a title or sequence.
; Nesting of sequences is not allowed.
;


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

	end

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
